//Technically this is provided by Winsock2.h
//https://msdn.microsoft.com/en-us/library/windows/desktop/ms740560%28v=vs.85%29.aspx?f=255&MSPPError=-2147217396
typedef struct timeval {
  long tv_sec;
  long tv_usec;
} timeval;
